package ru.littlevictor.radio;
import android.content.Context;
import android.view.LayoutInflater;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.widget.ArrayAdapter;
import android.widget.ImageView;
import android.widget.TextView;

import java.util.ArrayList;
import java.util.HashMap;

public class DynamicListAdapter extends ArrayAdapter<String> {

    final int INVALID_ID = -1;
    private LayoutInflater lInflater;
    private int theme[];

    HashMap<String, Integer> mIdMap = new HashMap<String, Integer>();

    public DynamicListAdapter(Context context, ArrayList<String> index) {
        super(context, 0, index);
        theme = ThemeChooser.getAppTheme(context);
        lInflater = (LayoutInflater) context.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
        for (int i = 0; i < index.size(); i++) {
            mIdMap.put(index.get(i), i);
        }
    };

    @Override
    public long getItemId(int position) {
        if (position < 0 || position >= mIdMap.size()) {
            return INVALID_ID;
        }
        String item = getItem(position);
        return mIdMap.get(item);
    };

    @Override
    public boolean hasStableIds() {
        return true;
    };

    @Override
    public View getView(int position, View view, ViewGroup parent) {
        if (view == null) {
            view = lInflater.inflate(R.layout.radio_edit_two_line_cell, parent, false);
        }
        ((TextView) view.findViewById(R.id.text1)).setText(RadioEdit.data.names.get(position));
        ((TextView) view.findViewById(R.id.text2)).setText(RadioEdit.data.urls.get(position));
        ImageView dragImg = (ImageView) view.findViewById(R.id.drag_image);
        dragImg.setOnTouchListener(new View.OnTouchListener() {
            @Override
            public boolean onTouch(View v, MotionEvent event) {
                if ((event.getAction() & MotionEvent.ACTION_MASK) == MotionEvent.ACTION_DOWN) DynamicListView.dragIsOn = true;
                return false;
            };
        });
        if(theme[0] == R.color.color_theme_primary_1) {
            dragImg.setBackgroundResource(R.drawable.cell_drag_light);
            if ((position % 2) == 0 ) {
                view.setBackgroundResource(R.drawable.background_activated_list_odd_dark);
            } else {
                view.setBackgroundResource(R.drawable.background_activated_list_even_dark);
            }
        } else {
            if ((position % 2) == 0 ) {
                view.setBackgroundResource(R.drawable.background_activated_list_odd_light);
            } else {
                view.setBackgroundResource(R.drawable.background_activated_list_even_light);
            }
        }
        return view;
    };
};
