package ru.littlevictor.radio;

import android.app.Notification;
import android.app.PendingIntent;
import android.app.Service;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.content.res.Resources;
import android.media.AudioManager;
import android.media.AudioManager.OnAudioFocusChangeListener;
import android.media.MediaMetadataRetriever;
import android.media.MediaPlayer;
import android.media.MediaPlayer.OnCompletionListener;
import android.media.MediaPlayer.OnErrorListener;
import android.media.MediaPlayer.OnPreparedListener;
import android.media.RemoteControlClient;
import android.media.SoundPool;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.net.Uri;
import android.net.wifi.WifiManager;
import android.net.wifi.WifiManager.WifiLock;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.os.PowerManager;
import android.os.PowerManager.WakeLock;
import android.preference.PreferenceManager;
import android.support.v4.app.NotificationCompat;
import android.telephony.TelephonyManager;
import android.view.Gravity;
import android.view.KeyEvent;
import android.widget.RemoteViews;
import android.widget.Toast;

import org.videolan.libvlc.EventHandler;
import org.videolan.libvlc.LibVLC;
import org.videolan.libvlc.LibVlcException;
import org.videolan.libvlc.Media;

import java.io.IOException;
import java.lang.ref.WeakReference;

import ru.littlevictor.radio.WidgetBtn.WidgetBtnUpdateService;
import ru.littlevictor.radio.WidgetMain.WidgetUpdateService;

//-----------------------------------------------------------------------------

public class RadioService extends Service implements OnAudioFocusChangeListener {
	
	public final static int STATE = 0;
	public final static int NEXT = 1;
	public final static int PREV = 2;
	public final static int LOAD = 3;
	public final static int SAVE = 4;
	public final static int PLAY = 5;
	public final static int POS = 6;
	public final static int STOP = 7;
	public final static int EXT = 8;
	public final static int ERROR = 9;
	public final static int TERM = 10;
	public final static int SHORT = 11;
	private final static int TRACK = 12;
	private int NOTIFY_ID = 101;
	private static int serviceState = TERM;
	private int currentPosition = -1;
    private String trackName = "";
    private String trackNameAndInfo = "";
	private String defaultAudioPlayer = "0";
	private String sheme;
	private boolean resumePlay = false;
	private boolean headsetFirstHandle = false;
	private boolean sendMyStateOnDestroy = true;
	private boolean renewList = true;
    private boolean internal = false;
    public static boolean headsetBeepEnable = false;

	private Data data;
	private LibVLC mLibVLC;
	private MediaPlayer mPlayer;
	private Handler libVlcHandler = new LibVlcHandler(this);
    @SuppressWarnings("deprecation")
    private static RemoteControlClient client;
	private SharedPreferences sharedPreferences;
    private BroadcastReceiver broadcastEventReceiver;
	private BluetoothAdapter mBluetoothAdapter;
	private static Intent playStopIntent;
	private static Intent prevIntent;
	private static Intent nextIntent;
	private static Intent quitIntent;
	private WakeLock wakeLock;
	private WifiLock wifiLock;
    private static SoundPool soundPool;
    private static int sound1;
    private static int sound2;
    private static int sound3;
    private static int sound4;

	// -----------------------------------------------------------------------------
	@SuppressWarnings("deprecation")
    @Override
	public void onCreate() {
		super.onCreate();
	
		mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
		playStopIntent = new Intent(getResources().getString(R.string.media_play_stop));
		prevIntent = new Intent(getResources().getString(R.string.media_previous));
		nextIntent = new Intent(getResources().getString(R.string.media_next));
		quitIntent = new Intent(getResources().getString(R.string.media_close));
		
		sharedPreferences = PreferenceManager.getDefaultSharedPreferences(this);
		PowerManager powerManager = (PowerManager)this.getSystemService(Context.POWER_SERVICE);
        if(powerManager != null) {
            wakeLock = powerManager.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK,
                    "ru.littlevictor.radio.WakeLock");
            wakeLock.acquire();
        }
		WifiManager wifiManager = (WifiManager) getSystemService(Context.WIFI_SERVICE);
        if(wifiManager != null) {
            wifiLock = wifiManager.createWifiLock(WifiManager.WIFI_MODE_FULL,
                    "ru.littlevictor.radio.WifiLock");
            wifiLock.acquire();
        }
        MediaButtonReceiver.registerMediaButton(this);
	};

    // -----------------------------------------------------------------------------
    private void createBroadcastEventListener() {
        if(broadcastEventReceiver == null) {
            broadcastEventReceiver = new BroadcastReceiver() {
                public void onReceive(Context context, Intent intent) {
                    String action = intent.getAction();
                    if (action.equals(getResources().getString(R.string.media_play_stop))) {
                        resumePlay = false;
                        switch (serviceState) {
                            case LOAD:
                            case PLAY:
                                stop(false);
                                break;
                            case STOP:
                            case ERROR:
                                play();
                                break;
                        }
                    } else if (action.equals(getResources().getString(R.string.media_previous))) {
                        goPrev();
                    } else if (action.equals(getResources().getString(R.string.media_next))) {
                        goNext();
                    } else if (action.equals(getResources().getString(R.string.media_close))) {
                        stop(true);
                    } else if (action.equals(Intent.ACTION_HEADSET_PLUG)) {
                        if (headsetFirstHandle) {
                            if (intent.getIntExtra("state", 0) == 0) {
                                switch (serviceState) {
                                    case LOAD:
                                    case PLAY:
                                    case ERROR:
                                        if(sharedPreferences.getBoolean("headset_unplug_stop", false)) {
                                            stop(true);
                                        }
                                        break;
                                }
                            }
                        } else {
                            headsetFirstHandle = true;
                        }
                    } else if (action.equals(BluetoothDevice.ACTION_ACL_DISCONNECTED)) {
                        switch (serviceState) {
                            case LOAD:
                            case PLAY:
                            case ERROR:
                                if(sharedPreferences.getBoolean("a2dp_disconnect_stop", false)) {
                                    stop(true);
                                }
                                break;
                        }
                    }
                };
            };
            IntentFilter filter = new IntentFilter();
            filter.addAction(getResources().getString(R.string.media_play_stop));
            filter.addAction(getResources().getString(R.string.media_previous));
            filter.addAction(getResources().getString(R.string.media_next));
            filter.addAction(getResources().getString(R.string.media_close));
            filter.addAction(Intent.ACTION_HEADSET_PLUG);
            filter.setPriority(IntentFilter.SYSTEM_HIGH_PRIORITY);
            if (mBluetoothAdapter != null) filter.addAction(BluetoothDevice.ACTION_ACL_DISCONNECTED);
            registerReceiver(broadcastEventReceiver, filter);
        }
    };

	// -----------------------------------------------------------------------------
	@Override
	public int onStartCommand(Intent intent, int flags, int startId) {
		if(intent.hasExtra("internal")) {
			internal = true;
		} else {
			internal = false;
		}
		String path;
		int inputAction = intent.getIntExtra(getResources().getString(R.string.service_state), STATE);
		
		switch (inputAction) {
		
		case SHORT:
			Data checkData = new Data(this, Storage.getPath(this, Storage.MAIN_LIST));
			int checkPosition = intent.getIntExtra(getResources().getString(R.string.media_position), -1);
			if(checkPosition < 0 || checkPosition >= checkData.size()) {
				showToast(getResources().getString(R.string.item_not_found));
			} else {
				renewList = true;
				data = checkData;
				currentPosition = checkPosition;
				sharedPreferences.edit()
				.putString("last_playlist_path", data.getPath())
				.putInt("current_position", currentPosition)
				.apply();
                createBroadcastEventListener();
				play();
			}
			break;
		case LOAD:
			data = new Data(this, intent.getStringExtra(getResources().getString(R.string.data_path)));
			if(data.errorMessage.equalsIgnoreCase("")) {
				renewList = true;
				sharedPreferences.edit().putString("last_playlist_path", data.getPath()).apply();
				currentPosition = 0;
				sharedPreferences.edit().putInt("current_position", currentPosition).apply();
				stopSelf();
			} else {
				showToast(data.errorMessage);
			}
			break;
		case STATE:
			renewList = true;
			path = sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST));
			data = new Data(this, path);
			if(!data.errorMessage.equalsIgnoreCase("")) {
				if(path.equalsIgnoreCase("")) {
					sharedPreferences.edit().putString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)).apply();
					data.writeList();
				}
			}
			currentPosition = sharedPreferences.getInt("current_position", -1);
			if(currentPosition < 0 || currentPosition >= data.size()) {
				currentPosition = 0;
				sharedPreferences.edit().putInt("current_position", currentPosition).apply();
			}
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
                    NOTIFY_ID++;
                    setNotification(TRACK);
					sendMyState();
				break;
				case TERM:
					stopSelf();
				break;
			}
			break;
		case EXT:
			renewList = false;
			currentPosition = intent.getIntExtra(getResources().getString(R.string.media_position), -1);
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
					if(currentPosition < 0 || currentPosition >= data.size()) {
						sendMyStateOnDestroy = false;
						stopSelf();
					} else {
						sendMyStateOnDestroy = true;
						sharedPreferences.edit().putInt("current_position", currentPosition).apply();
						openInExternalApp(Uri.parse(data.urls.get(currentPosition)));
					}
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					if(currentPosition < 0 || currentPosition >= data.size()) {
						sendMyStateOnDestroy = false;
						stopSelf();
					} else {
						sendMyStateOnDestroy = true;
						sharedPreferences.edit().putInt("current_position", currentPosition).apply();
						openInExternalApp(Uri.parse(data.urls.get(currentPosition)));
					}
				break;
			}
			break;
		case PLAY:
			renewList = false;
			switch (serviceState) {
				case LOAD:
				case STOP:
					play();
				break;
				case ERROR:
				case PLAY:
					stop(true);
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					currentPosition = sharedPreferences.getInt("current_position", -1);
					if(currentPosition < 0 || currentPosition >= data.size()) {
						currentPosition = 0;
						sharedPreferences.edit().putInt("current_position", currentPosition).apply();
					}
                    createBroadcastEventListener();
					play();
				break;
			}
			break;
		case POS:
			renewList = false;
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
					currentPosition = intent.getIntExtra(getResources().getString(R.string.media_position), -1);
					if(currentPosition < 0 || currentPosition >= data.size()) {
						currentPosition = sharedPreferences.getInt("current_position", -1);
						if(currentPosition < 0 || currentPosition >= data.size()) {
							currentPosition = 0;
							sharedPreferences.edit().putInt("current_position", currentPosition).apply();
						}
					}
					play();
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					currentPosition = intent.getIntExtra(getResources().getString(R.string.media_position), -1);
					if(currentPosition < 0 || currentPosition >= data.size()) {
						currentPosition = sharedPreferences.getInt("current_position", -1);
						if(currentPosition < 0 || currentPosition >= data.size()) {
							currentPosition = 0;
							sharedPreferences.edit().putInt("current_position", currentPosition).apply();
						}
					}
                    createBroadcastEventListener();
					play();
				break;
			}
			break;
		case NEXT:
			renewList = false;
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
					goNext();
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					currentPosition = sharedPreferences.getInt("current_position", -1);
					if(currentPosition < 0 || currentPosition >= data.size()) {
						currentPosition = 0;
						sharedPreferences.edit().putInt("current_position", currentPosition).apply();
					}
					goNext();
					stopSelf();
				break;
			}
			break;
		case PREV:
			renewList = false;
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
					goPrev();
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					currentPosition = sharedPreferences.getInt("current_position", -1);
					if(currentPosition < 0 || currentPosition >= data.size()) {
						currentPosition = 0;
						sharedPreferences.edit().putInt("current_position", currentPosition).apply();
					}
					goPrev();
					stopSelf();
				break;
			}
			break;
		case SAVE:
			renewList = false;
			path = intent.getStringExtra(getResources().getString(R.string.data_path));
			switch (serviceState) {
				case LOAD:
				case PLAY:
				case ERROR:
				case STOP:
					data.writeList(path);
					if(!data.errorMessage.equalsIgnoreCase("")) {
						showToast(data.errorMessage);
					}
				break;
				case TERM:
					data = new Data(this, sharedPreferences.getString("last_playlist_path", Storage.getPath(this, Storage.MAIN_LIST)));
					data.writeList(path);
					if(!data.errorMessage.equalsIgnoreCase("")) {
						showToast(data.errorMessage);
					}
					sendMyStateOnDestroy = false;
					stopSelf();
				break;
			}
			break;
		}
		
		return START_NOT_STICKY;
	};

	// -----------------------------------------------------------------------------
	private void sendMyState() {
        boolean sendBitrate = sharedPreferences.getBoolean("show_bitrate", false);
		sendBroadcast(new Intent(getResources().getString(R.string.service_out_action))
		.putExtra(getResources().getString(R.string.service_state), serviceState)
		.putExtra(getResources().getString(R.string.media_position), currentPosition)
		.putExtra(getResources().getString(R.string.media_track), sendBitrate?trackNameAndInfo:trackName )
		.putExtra(getResources().getString(R.string.data_path), data.getPath())
		.putExtra(getResources().getString(R.string.reload_list), renewList)
		.putStringArrayListExtra(getResources().getString(R.string.media_name), data.names)
		.putStringArrayListExtra(getResources().getString(R.string.media_url), data.urls));

        if(Build.VERSION.SDK_INT != 20) {
            startService(new Intent(this, WidgetUpdateService.class)
                    .putExtra(getResources().getString(R.string.service_state), serviceState)
                    .putExtra(getResources().getString(R.string.media_position), currentPosition)
                    .putExtra(getResources().getString(R.string.media_track), trackName)
                    .putStringArrayListExtra(getResources().getString(R.string.media_name), data.names)
                    .putStringArrayListExtra(getResources().getString(R.string.media_url), data.urls));

            startService(new Intent(this, WidgetBtnUpdateService.class)
                    .putExtra(getResources().getString(R.string.service_state), serviceState));
        }
	};

	// -----------------------------------------------------------------------------
	private void setNotification(int notificationType){
		int[] theme = ThemeChooser.getNotifyTheme(this);
		int notificationBgRes = theme[0];
		int themeLayoutRes = theme[1];
		int themeLayoutResBig = theme[2];
		Resources res = getResources();
		RemoteViews smallContentView = new RemoteViews(getPackageName(), themeLayoutRes);
		RemoteViews bigContentView = new RemoteViews(getPackageName(), themeLayoutResBig);
      
		Intent notificationIntent = new Intent(this, RadioMain.class);
		notificationIntent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_CLEAR_TASK);
		PendingIntent contentIntent = PendingIntent.getActivity(getBaseContext(), 0, notificationIntent, 0);
		
		NotificationCompat.Builder builder = new NotificationCompat.Builder(this);

		switch (notificationType) {
			case TRACK:
				builder.setContentIntent(contentIntent).setSmallIcon(R.drawable.ic_stat_launcher);
		        bigContentView.setTextViewText(R.id.track_name_container, trackName);
		        bigContentView.setImageViewResource(R.id.notifyButtonPlayStop, R.drawable.ic_media_pause);
		        bigContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
		        smallContentView.setTextViewText(R.id.track_name_container, trackName);
		        smallContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
			break;
			case LOAD:
				serviceState = LOAD;
				builder.setContentIntent(contentIntent).setSmallIcon(android.R.drawable.ic_popup_sync);
				bigContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        bigContentView.setImageViewResource(R.id.notifyButtonPlayStop, R.drawable.ic_media_pause);
		        bigContentView.setImageViewResource(R.id.notifyBigImage, android.R.drawable.ic_popup_sync);
				smallContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        smallContentView.setImageViewResource(R.id.notifyBigImage, android.R.drawable.ic_popup_sync);
			break;
			case PLAY:
				serviceState = PLAY;
				builder.setContentIntent(contentIntent).setSmallIcon(R.drawable.ic_stat_launcher);
		        bigContentView.setTextViewText(R.id.track_name_container, trackName);
		        bigContentView.setImageViewResource(R.id.notifyButtonPlayStop, R.drawable.ic_media_pause);
		        bigContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
		        smallContentView.setTextViewText(R.id.track_name_container, trackName);
		        smallContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
			break;
			case STOP:
				serviceState = STOP;
				builder.setContentIntent(contentIntent).setSmallIcon(R.drawable.ic_stat_launcher);
				bigContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        bigContentView.setImageViewResource(R.id.notifyButtonPlayStop, R.drawable.ic_media_play);
		        bigContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
				smallContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        smallContentView.setImageViewResource(R.id.notifyBigImage, R.drawable.ic_stat_launcher);
			break;
			case ERROR:
				serviceState = ERROR;
				builder.setContentIntent(contentIntent).setSmallIcon(android.R.drawable.stat_sys_warning);
				builder.setContentIntent(contentIntent).setTicker(res.getString(R.string.ticker_error) + " " + data.names.get(currentPosition));
				bigContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        bigContentView.setImageViewResource(R.id.notifyButtonPlayStop, R.drawable.ic_media_play);
		        bigContentView.setImageViewResource(R.id.notifyBigImage, android.R.drawable.stat_sys_warning);
				smallContentView.setTextViewText(R.id.track_name_container, res.getString(R.string.no_name));
		        smallContentView.setImageViewResource(R.id.notifyBigImage, android.R.drawable.stat_sys_warning);
			break;
		};
      
		if(notificationBgRes > 0) {
			bigContentView.setInt(R.id.content, "setBackgroundResource", notificationBgRes);
			smallContentView.setInt(R.id.content, "setBackgroundResource", notificationBgRes);
		}
        String notifyTitle = data.names.get(currentPosition);
		bigContentView.setTextViewText(R.id.notifyTitle, notifyTitle);
		smallContentView.setTextViewText(R.id.notifyTitle, notifyTitle);
		
		bigContentView.setOnClickPendingIntent(R.id.notifyButtonPrev, PendingIntent.getBroadcast(this, 0, prevIntent, 0));
		bigContentView.setOnClickPendingIntent(R.id.notifyButtonPlayStop, PendingIntent.getBroadcast(this, 0, playStopIntent, 0));
		bigContentView.setOnClickPendingIntent(R.id.notifyButtonNext, PendingIntent.getBroadcast(this, 0, nextIntent, 0));
		bigContentView.setOnClickPendingIntent(R.id.notifyButtonQuit, PendingIntent.getBroadcast(this, 0, quitIntent, 0));
		smallContentView.setOnClickPendingIntent(R.id.notifyButtonQuit, PendingIntent.getBroadcast(this, 0, quitIntent, 0));
		
		sharedPreferences.edit().putInt("current_position", currentPosition).apply();
		
		if(Build.VERSION.SDK_INT >= 16) {
			Notification n = builder.build();
			n.contentView = smallContentView;
			n.bigContentView = bigContentView;
			n.flags = n.flags | Notification.FLAG_INSISTENT;
			startForeground (NOTIFY_ID, n);
		} else {
            Notification n = builder.build();
            n.contentView = smallContentView;
            n.flags = n.flags | Notification.FLAG_INSISTENT;
            startForeground (NOTIFY_ID, n);
		};

        updateRemoteControlClient(notifyTitle);
		sendMyState();
	};
	
	// -----------------------------------------------------------------------------
	private static boolean isOnline(Context context)
	{
		ConnectivityManager cm = (ConnectivityManager) context.getSystemService(Context.CONNECTIVITY_SERVICE);
		NetworkInfo netInfo = cm.getActiveNetworkInfo();
		if (netInfo != null && netInfo.isConnectedOrConnecting())
		{
		   return true;
		}
		return false;
	};
	
	// -----------------------------------------------------------------------------
	private void play() {
        headsetBeepEnable = sharedPreferences.getBoolean("headset_beep_enable", false);
        if(headsetBeepEnable && soundPool == null) {
            soundPool = new SoundPool(1, AudioManager.STREAM_MUSIC, 0);
            sound1 = soundPool.load(this, R.raw.beep1, 1);
            sound2 = soundPool.load(this, R.raw.beep2, 1);
            sound3 = soundPool.load(this, R.raw.beep3, 1);
            sound4 = soundPool.load(this, R.raw.beep4, 1);
        }
        defaultAudioPlayer = sharedPreferences.getString("default_audio_player", "0");
        // -----------------------------------------------------------------------------
        resumePlay = false;
        trackNameAndInfo = getResources().getString(R.string.no_name);
        String lastURL = data.urls.get(currentPosition);
        Uri uri;
        String path;
        // -----------------------------------------------------------------------------
        try
        {
            uri = Uri.parse(lastURL);
            sheme = uri.getScheme();
            path = uri.getPath().toLowerCase();
        }
        catch (Exception e)
        {
            showToast(getResources().getString(R.string.invalid_url));
            return;
        };
        // -----------------------------------------------------------------------------
        if (uri == null || sheme == null || path == null)
        {
            showToast(getResources().getString(R.string.invalid_url));
            return;
        };
        if(!sheme.equalsIgnoreCase("file")){
            if(!isOnline(this)) {
                Intent intent = new Intent("android.intent.action.MAIN")
                        .setClass(this, DialogNetwork.class)
                        .setFlags(Intent.FLAG_ACTIVITY_NEW_TASK
                                | Intent.FLAG_ACTIVITY_NO_HISTORY
                                | Intent.FLAG_ACTIVITY_RETAIN_IN_RECENTS);
                startActivity(intent);
                return;
            }
        }
        // -----------------------------------------------------------------------------
        if(path.endsWith(".m3u") ||
           path.endsWith(".pls") ||
           path.endsWith(".m3u8") ||
           path.endsWith(".ram")) {
            if(sheme.equalsIgnoreCase("http")) {
                if(internal) ListStack.push(getApplicationContext(), data.getPath());
                startActivity(new Intent(this, RadioDownloader.class)
                .putExtra("internal", true)
                .setData(uri)
                .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_NO_HISTORY));
            } else if(sheme.equalsIgnoreCase("file")) {
                if(internal) {
                    ListStack.push(getApplicationContext(), data.getPath());
                    Data checkData = new Data(this, uri.getPath());
                    if(checkData.errorMessage.equalsIgnoreCase("")) {
                        data = checkData;
                        renewList = true;
                        sharedPreferences.edit().putString("last_playlist_path", data.getPath()).apply();
                        currentPosition = 0;
                        sharedPreferences.edit().putInt("current_position", currentPosition).apply();
                        stopSelf();
                    } else {
                        showToast(checkData.errorMessage);
                    }
                } else {
                    startActivity(new Intent(this, RadioServiceActivity.class)
                    .setData(uri)
                    .setFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_NO_HISTORY));
                }
            } else {
                openInExternalApp(uri);
            }
        } else if(sheme.equalsIgnoreCase("http") ||
                  sheme.equalsIgnoreCase("rtsp") ||
                  sheme.equalsIgnoreCase("mms") ||
                  sheme.equalsIgnoreCase("file")) {
            if (defaultAudioPlayer.equalsIgnoreCase("0")) {
                playToVLC(lastURL);
            } else if(defaultAudioPlayer.equalsIgnoreCase("1")) {
                playToAndroid(lastURL);
            } else if(defaultAudioPlayer.equalsIgnoreCase("2")) {
                openInExternalApp(Uri.parse(lastURL));
            }
        } else if(sheme.equalsIgnoreCase("udp")) {
            sendMyStateOnDestroy = true;
            sharedPreferences.edit().putInt("current_position", currentPosition).apply();
            stop(true);
            Intent intent = new Intent().setAction(Intent.ACTION_VIEW).setData(uri)
            .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_NO_HISTORY | Intent.FLAG_ACTIVITY_CLEAR_TASK);
            if (intent.resolveActivity(getPackageManager()) != null) {
                startActivity(intent);
            } else {
                showToast(getResources().getString(R.string.activity_not_resolve));
            }
        }
	};

    // -----------------------------------------------------------------------------
	private void playToVLC(String lastURL) {
        setNotification(LOAD);
        releaseMediaPlayer();
        int audioOutput;
        int bufferingTime;
        String vlcAudioOutputValue = PreferenceManager.getDefaultSharedPreferences(this).getString("vlc_audio_output", "");
        if(vlcAudioOutputValue.equalsIgnoreCase("1")) {
            audioOutput = 1;
        } else {
            audioOutput = 0;
        }
        try {
            bufferingTime = Integer.parseInt(PreferenceManager.getDefaultSharedPreferences(this).getString("vlc_buffering_time", ""));
        } catch(NumberFormatException nfe) {
            bufferingTime = 0;
        }
        try {
            if (mLibVLC == null) {
                mLibVLC = LibVLC.getInstance();
                mLibVLC.init(RadioService.this);
                libVlcHandler = new LibVlcHandler(this);
                EventHandler.getInstance().addHandler(libVlcHandler);
                mLibVLC.setAout(audioOutput);
                mLibVLC.setNetworkCaching(bufferingTime);
                mLibVLC.playMRL(lastURL);
                mLibVLC.setVolume(100);
            } else {
                mLibVLC.setAout(audioOutput);
                mLibVLC.setNetworkCaching(bufferingTime);
                mLibVLC.playMRL(lastURL);
                mLibVLC.setVolume(100);
            };
        } catch(LibVlcException e) {
            showToast(getResources().getString(R.string.libvlc_error));
            releaseVlcPlayer();
            setNotification(ERROR);
            return;
        }
        ((AudioManager) getSystemService(AUDIO_SERVICE)).requestAudioFocus(this,
                AudioManager.STREAM_MUSIC, AudioManager.AUDIOFOCUS_GAIN);
    };

    // -----------------------------------------------------------------------------
    private void playToAndroid(String lastURL) {
        setNotification(LOAD);
        releaseVlcPlayer();
        releaseMediaPlayer();
        trackNameAndInfo = trackName = getResources().getString(R.string.no_name);
        mPlayer = new MediaPlayer();
        mPlayer.setVolume(1.0f, 1.0f);
        mPlayer.setOnErrorListener(new OnErrorListener() {
            @Override
            public boolean onError(MediaPlayer mp, int what, int extra) {
                setNotification(ERROR);
                if (mPlayer != null) {
                    try {
                        mPlayer.release();
                    } finally {
                        mPlayer = null;
                    }
                }
                if(headsetBeepEnable) beep(4);
                return false;
            }
        });
        mPlayer.setOnPreparedListener(new OnPreparedListener() {
            @Override
            public void onPrepared(MediaPlayer mp) {
                mp.start();
                setNotification(PLAY);
            }
        });
        mPlayer.setOnCompletionListener(new OnCompletionListener() {
            @Override
            public void onCompletion(MediaPlayer mp) {
                goNext();
            }
        });
        try {
            mPlayer.setDataSource(lastURL);
            mPlayer.prepareAsync();
        } catch (IOException e) {
            setNotification(ERROR);
        } catch (IllegalArgumentException e) {
            setNotification(ERROR);
        }
        ((AudioManager) getSystemService(AUDIO_SERVICE)).requestAudioFocus(this,
                AudioManager.STREAM_MUSIC, AudioManager.AUDIOFOCUS_GAIN);
    };

    // -----------------------------------------------------------------------------
    @SuppressWarnings("deprecation")
    private void updateRemoteControlClient(String title) {
        if(client != null) {
            client.editMetadata(true)
                    .putString(MediaMetadataRetriever.METADATA_KEY_TITLE, trackName + " " + title)
                    .apply();
            client.setPlaybackState(RemoteControlClient.PLAYSTATE_FAST_FORWARDING);
            client.setPlaybackState(RemoteControlClient.PLAYSTATE_REWINDING);
            switch(serviceState) {
                case LOAD:
                    client.setPlaybackState(RemoteControlClient.PLAYSTATE_BUFFERING);
                    break;
                case PLAY:
                    client.setPlaybackState(RemoteControlClient.PLAYSTATE_PLAYING);
                    break;
                case STOP:
                    client.setPlaybackState(RemoteControlClient.PLAYSTATE_STOPPED);
                    break;
                case ERROR:
                    client.setPlaybackState(RemoteControlClient.PLAYSTATE_ERROR);
                    break;
            }
        }
    };

	// -----------------------------------------------------------------------------
	private void goPrev() {
        trackNameAndInfo = trackName = getResources().getString(R.string.no_name);
		if(data.size() > 1) {
			if(currentPosition == 0) {
				currentPosition = data.size() - 1;
			} else {
				currentPosition--;
			}
			sharedPreferences.edit().putInt("current_position", currentPosition).apply();
		}
        internal = true;
        switch (serviceState) {
            case LOAD:
            case PLAY:
            case ERROR:
                play();
                break;
            case STOP:
                stop(false);
                break;
        }
        internal = false;
	};

	// -----------------------------------------------------------------------------
	private void goNext() {
        trackNameAndInfo = trackName = getResources().getString(R.string.no_name);
		if(data.size() > 1) {
			if(currentPosition == data.size() - 1) {
				currentPosition = 0;
			} else {
				currentPosition++;
			}
			sharedPreferences.edit().putInt("current_position", currentPosition).apply();
		}
        internal = true;
        switch (serviceState) {
            case LOAD:
            case PLAY:
            case ERROR:
                play();
                break;
            case STOP:
                stop(false);
                break;
        }
        internal = false;
	};

	// -----------------------------------------------------------------------------
	private void openInExternalApp(Uri uri) {
		stop(true);
		String sheme;
		String path;
		// -----------------------------------------------------------------------------
	    try 
	    {
			sheme = uri.getScheme();
			path = uri.getPath();
	    }
	    catch (Exception e) 
	    {
			showToast(getResources().getString(R.string.invalid_url));
            return;
	    }
		// -----------------------------------------------------------------------------
	    if (uri == null || sheme == null || path == null) 
	    {
			showToast(getResources().getString(R.string.invalid_url));
            return;
	    };
		if(sheme.equalsIgnoreCase("udp")) {
			Intent i = new Intent().setAction(Intent.ACTION_VIEW).setData(uri)
					.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_NO_HISTORY | Intent.FLAG_ACTIVITY_CLEAR_TASK);
				    if (i.resolveActivity(getPackageManager()) != null) {
				        startActivity(i);
				    } else {
						showToast(getResources().getString(R.string.activity_not_resolve));
					}
		} else {
			Intent i = new Intent().setAction(Intent.ACTION_VIEW).setDataAndType(uri, "audio/mpeg")
					.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK | Intent.FLAG_ACTIVITY_NO_HISTORY | Intent.FLAG_ACTIVITY_CLEAR_TASK);
				    if (i.resolveActivity(getPackageManager()) != null) {
				        startActivity(i);
				    } else {
						showToast(getResources().getString(R.string.activity_not_resolve));
					}
		}
	};
	
	// -----------------------------------------------------------------------------
	public void stop(boolean cancelNotify) {
		trackNameAndInfo = trackName = getResources().getString(R.string.no_name);
        releaseMediaPlayer();
        if (mLibVLC != null) {
            mLibVLC.stop();
        }
		if(cancelNotify) {
			stopForeground(true);
			stopSelf();
		} else {
            setNotification(STOP);
		}
	};

	// -----------------------------------------------------------------------------
	@Override
	public void onDestroy() {
		super.onDestroy();
        ((AudioManager) getSystemService(AUDIO_SERVICE)).abandonAudioFocus(this);
        MediaButtonReceiver.unregisterMediaButton(this);
        if(broadcastEventReceiver != null) unregisterReceiver(broadcastEventReceiver);
		if(wakeLock != null) wakeLock.release();
        if(wifiLock != null) wifiLock.release();
        releaseMediaPlayer();
		releaseVlcPlayer();
		trackNameAndInfo = getResources().getString(R.string.no_name);
		serviceState = TERM;
		if(sendMyStateOnDestroy) sendMyState();
	};

	// -----------------------------------------------------------------------------
	@Override
	public void onAudioFocusChange(int focusChange) {
		switch (focusChange) {
		case AudioManager.AUDIOFOCUS_LOSS:
			resumePlay = true;
			stop(false);
		break;
		case AudioManager.AUDIOFOCUS_GAIN:
			if(resumePlay) {
				resumePlay = false;
				play();
			} else {
				if (mPlayer != null) {
					mPlayer.setVolume(1.0f, 1.0f);
				}
		        if (mLibVLC != null) {
                    mLibVLC.setVolume(100);
		        }
			}
		break;
		default:
			if (mPlayer != null) {
				mPlayer.setVolume(0.0f, 0.0f);
			}
	        if (mLibVLC != null) {
	    		mLibVLC.setVolume(0);
	        }
		break;
		}
	};

    // -----------------------------------------------------------------------------
    private static void beep(int id) {
        switch(id) {
            case 1:
                soundPool.play(sound1, 1.0f, 1.0f, 1000, 0, 1.0f);
                break;
            case 2:
                soundPool.play(sound2, 1.0f, 1.0f, 1000, 0, 1.0f);
                break;
            case 3:
                soundPool.play(sound3, 1.0f, 1.0f, 1000, 0, 1.0f);
                break;
            case 4:
                soundPool.play(sound4, 1.0f, 1.0f, 1000, 0, 1.0f);
                break;
        }
    };

    // -----------------------------------------------------------------------------
	@SuppressWarnings("deprecation")
    public static class MediaButtonReceiver extends BroadcastReceiver {
        private static long eventDownTime = 0;
        private static long eventUpTime = 0;
        private static int clickCount = 0;
        private static TelephonyManager telephonyManager;
        private static Context context;
        private static Runnable runnable = new Runnable() {
            @Override
            public void run() {
                switch (clickCount) {
                    case 1:
                        if(headsetBeepEnable) beep(1);
                        if(serviceState == RadioService.TERM) {
                            context.startService(new Intent(context, RadioService.class)
                                    .putExtra("internal", true)
                                    .putExtra(context.getResources().getString(R.string.service_state), RadioService.PLAY));
                        } else {
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_play_stop)));
                        }
                        break;
                    case 2:
                        if(headsetBeepEnable) beep(2);
                        if(serviceState == RadioService.TERM) {
                            context.startService(new Intent(context, RadioService.class)
                                    .putExtra("internal", true)
                                    .putExtra(context.getResources().getString(R.string.service_state), RadioService.NEXT));
                        } else {
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_next)));
                        }
                        break;
                    case 3:
                        if(headsetBeepEnable) beep(3);
                        if(serviceState == RadioService.TERM) {
                            context.startService(new Intent(context, RadioService.class)
                                    .putExtra("internal", true)
                                    .putExtra(context.getResources().getString(R.string.service_state), RadioService.PREV));
                        } else {
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_previous)));
                        }
                        break;
                }
                clickCount = 0;
            }
        };

		public MediaButtonReceiver() {
			super();
		};

        // -----------------------------------------------------------------------------
        public boolean processKey(Context context, KeyEvent event) {
            if ((telephonyManager != null?telephonyManager.getCallState() == TelephonyManager.CALL_STATE_IDLE:false) && !RadioMain.running) {
                int action = event.getAction();
                switch (event.getKeyCode()) {
                    case KeyEvent.KEYCODE_MEDIA_PLAY_PAUSE:
                    case KeyEvent.KEYCODE_MEDIA_PLAY:
                    case KeyEvent.KEYCODE_MUTE:
                    case KeyEvent.KEYCODE_MEDIA_PAUSE:
                    case KeyEvent.KEYCODE_MEDIA_STOP:
                        if (action == KeyEvent.ACTION_UP) {
//                            if(headsetBeepEnable) beep(1);
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_play_stop)));
                        }
                        return true;
                    case KeyEvent.KEYCODE_MEDIA_FAST_FORWARD:
                    case KeyEvent.KEYCODE_MEDIA_NEXT:
                        if (action == KeyEvent.ACTION_UP) {
//                            if(headsetBeepEnable) beep(2);
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_next)));
                        }
                        return true;
                    case KeyEvent.KEYCODE_MEDIA_REWIND:
                    case KeyEvent.KEYCODE_MEDIA_PREVIOUS:
                        if (action == KeyEvent.ACTION_UP) {
//                            if(headsetBeepEnable) beep(3);
                            context.sendBroadcast(new Intent(context.getResources().getString(R.string.media_previous)));
                        }
                        return true;
                    case KeyEvent.KEYCODE_HEADSETHOOK:
                        switch (action) {
                            case KeyEvent.ACTION_DOWN:
                                clickCount++;
                                eventDownTime = event.getDownTime();
                                break;
                            case KeyEvent.ACTION_UP:
                                eventUpTime = event.getEventTime();
                                if(eventDownTime == eventUpTime) {
                                    if(headsetBeepEnable) beep(1);
                                    if (clickCount == 1) {
                                        new Handler().postDelayed(runnable, 3000);
                                    }
                                } else {
                                    if (clickCount == 1) {
                                        new Handler().postDelayed(runnable, 500);
                                    }
                                }
                                break;
                        }
                        return true;
                }
            }
            return false;
        };

        // -----------------------------------------------------------------------------
        public static void registerMediaButton(Context context) {
            MediaButtonReceiver.context = context;
            telephonyManager = (TelephonyManager)context.getSystemService(Context.TELEPHONY_SERVICE);
            AudioManager audioManager = (AudioManager)context.getSystemService(Context.AUDIO_SERVICE);
            ComponentName receiver = new ComponentName(context.getPackageName(), MediaButtonReceiver.class.getName());
            audioManager.registerMediaButtonEventReceiver(receiver);
            if(Build.VERSION.SDK_INT < 20) {
                client = new RemoteControlClient(PendingIntent.getBroadcast(context, 0,
                        new Intent(Intent.ACTION_MEDIA_BUTTON).setComponent(receiver), 0));
                client.setTransportControlFlags(
                        RemoteControlClient.FLAG_KEY_MEDIA_PLAY_PAUSE |
                                RemoteControlClient.FLAG_KEY_MEDIA_PREVIOUS |
                                RemoteControlClient.FLAG_KEY_MEDIA_NEXT);
                audioManager.registerRemoteControlClient(client);
            }
        };

        // -----------------------------------------------------------------------------
        public static void unregisterMediaButton(Context context) {
            AudioManager audioManager = (AudioManager)context.getSystemService(Context.AUDIO_SERVICE);
            ComponentName receiver = new ComponentName(context.getPackageName(), MediaButtonReceiver.class.getName());
            if(client != null) audioManager.unregisterRemoteControlClient(client);
            audioManager.unregisterMediaButtonEventReceiver(receiver);
        };

        // -----------------------------------------------------------------------------
        @Override
        public void onReceive(Context context, Intent intent) {
            MediaButtonReceiver.context = context;
            if (Intent.ACTION_MEDIA_BUTTON.equals(intent.getAction())) {
                KeyEvent event = intent.getParcelableExtra(Intent.EXTRA_KEY_EVENT);
                boolean handled = processKey(context, event);
                if (handled && isOrderedBroadcast()) abortBroadcast();
            }
        };
	};

    // -----------------------------------------------------------------------------
	private class LibVlcHandler extends Handler {
		
		private WeakReference<RadioService> mOwner;
		private boolean endReachedFlag = false;
		public LibVlcHandler(RadioService owner) {
			mOwner = new WeakReference(owner);
		};

		@Override
		public void handleMessage(Message msg) {
			RadioService player = mOwner.get();
			Bundle b = msg.getData();
			switch (b.getInt("event")) {
                case EventHandler.MediaMetaChanged:
                    if(!player.sheme.equalsIgnoreCase("file")) {
                        String track = "";
                        String trackInfo = "";
                        String name = "";
                        String genre = "";
                        String title = "";
                        String info = "";
                        boolean genreExist = false;
                        boolean nameExist = false;
                        boolean titleExist = false;
                        boolean infoExist = false;
                        name = mLibVLC.getMeta(Media.Meta.Title);
                        genre = mLibVLC.getMeta(Media.Meta.Genre);
                        title = mLibVLC.getMeta(Media.Meta.NowPlaying);
                        // -----------------------------------------------------------------------------
                        if(name == null || name.length() < 3){
                            nameExist = false;
                        } else {
                            nameExist = true;
                            trackInfo = (titleExist?(track + "\n"):"")
                                    + (name + (genreExist?(" (" + genre + ")"):""))
                                    + (infoExist?("\n" + info):"");
                        }
                        // -----------------------------------------------------------------------------
                        if(genre == null || genre.length() < 3){
                            genreExist = false;
                        } else {
                            genreExist = true;
                            trackInfo = (titleExist?(track + "\n"):"")
                                    + (nameExist?(name + " "):"") + "(" + genre + ")"
                                    + (infoExist?("\n" + info):"");
                        }
                        // -----------------------------------------------------------------------------
                        if(title == null || title.length() < 5){
                            titleExist = false;
                            track = "";
                        } else {
                            titleExist = true;
                            track = title;
                            trackInfo = track + "\n"
                                    + (nameExist?(name + (genreExist?(" (" + genre + ")"):"")):"")
                                    + (infoExist?("\n" + info):"");
                        }
                        // -----------------------------------------------------------------------------
                        if(serviceState == PLAY) {
                            if(!track.equals(player.trackName) || !trackInfo.equals(player.trackNameAndInfo)) {
                                player.renewList = false;
                                player.trackName = track;
                                player.trackNameAndInfo = trackInfo;
                                player.setNotification(TRACK);
                            }
                        }
                    }
                    break;
				case EventHandler.MediaPlayerEndReached:
					endReachedFlag = true;
					player.mLibVLC.stop();
					break;
				case EventHandler.MediaPlayerPlaying:
					player.setNotification(PLAY);
					break;
				case EventHandler.MediaPlayerPaused:
				case EventHandler.MediaPlayerStopped:
					if(endReachedFlag) {
						endReachedFlag = false;
						if(player.sheme.equalsIgnoreCase("file")) {
							player.goNext();
						} else {
							player.playAsStream();
						}
					} else {
                        if (player.defaultAudioPlayer.equalsIgnoreCase("0")) {
                            player.setNotification(STOP);
                        }
					}
					break;
				case EventHandler.MediaPlayerEncounteredError:
					endReachedFlag = false;
					player.setNotification(ERROR);
					player.mLibVLC.stop();
                    if(headsetBeepEnable) beep(4);
					break;
			}
		};
	};

	// -----------------------------------------------------------------------------
	private void playAsStream() {
		mLibVLC.playMRL(data.urls.get(currentPosition) + "/;");
	};

	// -----------------------------------------------------------------------------
	private void releaseVlcPlayer() {
		if (mLibVLC != null) {
			EventHandler.getInstance().removeHandler(libVlcHandler);
			mLibVLC.stop();
			mLibVLC.destroy();
			mLibVLC = null;
		}
	};

    // -----------------------------------------------------------------------------
    private void releaseMediaPlayer() {
        if (mPlayer != null) {
            try {
                mPlayer.release();
            } finally {
                mPlayer = null;
            }
        }
    };

	// -----------------------------------------------------------------------------
	private void showToast(String message) {
        int[] theme = ThemeChooser.getAppTheme(this);
		Toast toast = Toast.makeText(this, message, Toast.LENGTH_LONG);
		toast.setGravity(Gravity.CENTER_HORIZONTAL | Gravity.CENTER_VERTICAL, 0, 0);
        if(theme[0] == R.color.color_theme_primary_1) {
            toast.getView().setBackgroundResource(R.drawable.background_toast_dark);
        } else {
            toast.getView().setBackgroundResource(R.drawable.background_toast_light);
        }
		toast.show();
	};

    // -----------------------------------------------------------------------------
	public static int getState() {
        return serviceState;
    };

	// -----------------------------------------------------------------------------
	@Override
	public IBinder onBind(Intent intent) {
		return null;
	};
 
	// -----------------------------------------------------------------------------
};